@@ -29,7 +29,7 @@ import subprocess
 
 import libcalamares
 
-from libcalamares.utils import check_target_env_call, check_target_env_output
+from libcalamares.utils import target_env_call, check_target_env_call, check_target_env_output
 
 import gettext
 
@@ -62,6 +62,12 @@ def get_uuid():
 
     return ""
 
+def get_machine_id(installation_root_path):
+    with open(os.path.join(installation_root_path, "etc", "machine-id"), 'r') as machineid_file:
+        machine_id = machineid_file.read().rstrip('\n')
+
+    return machine_id
+
 
 def get_kernel_line(kernel_type):
     """
@@ -231,8 +237,7 @@ def create_systemd_boot_conf(installation_root_path, efi_dir, uuid, kernel, kern
     libcalamares.utils.debug(f"Configuring kernel version {kernel_version}")
 
     # get the machine-id
-    with open(os.path.join(installation_root_path, "etc", "machine-id"), 'r') as machineid_file:
-        machine_id = machineid_file.read().rstrip('\n')
+    machine_id = get_machine_id(installation_root_path)
 
     # Ensure the directory exists
     machine_dir = os.path.join(installation_root_path + efi_dir, machine_id)
@@ -396,7 +401,7 @@ def get_efi_suffix_generator(name):
     if name.count("${") > 1:
         raise ValueError("EFI ID {!r} contains multiple generators".format(name))
     import re
-    prefix, generator_name = re.match("(.*)\${([^}]*)}$", name).groups()
+    prefix, generator_name = re.match(r"(.*)\${([^}]*)}$", name).groups()
     if generator_name not in ("SERIAL", "RANDOM", "PHRASE"):
         raise ValueError("EFI suffix {!r} is unknown".format(generator_name))
 
@@ -623,7 +628,6 @@ def run_grub_install(fw_type, partitions, efi_directory, install_hybrid_grub):
     """
 
     is_zfs = any([is_zfs_root(partition) for partition in partitions])
-
     # zfs needs an environment variable set for grub
     if is_zfs:
         check_target_env_call(["sh", "-c", "echo ZPOOL_VDEV_NAME_PATH=1 >> /etc/environment"])
@@ -633,22 +637,21 @@ def run_grub_install(fw_type, partitions, efi_directory, install_hybrid_grub):
         efi_bootloader_id = efi_label(efi_directory)
         efi_target, efi_grub_file, efi_boot_file = get_grub_efi_parameters()
 
-        if is_zfs:
-            check_target_env_call(["sh", "-c", "ZPOOL_VDEV_NAME_PATH=1 " + libcalamares.job.configuration["grubInstall"]
-                                   + " --target=" + efi_target + " --efi-directory=" + efi_directory
-                                   + " --bootloader-id=" + efi_bootloader_id + " --force"])
-        else:
-            check_target_env_call([libcalamares.job.configuration["grubInstall"],
-                                   "--target=" + efi_target,
-                                   "--efi-directory=" + efi_directory,
-                                   "--bootloader-id=" + efi_bootloader_id,
-                                   "--force"])
-    else:
-        if efi_directory is not None and not install_hybrid_grub:
-            libcalamares.utils.warning(_("Cannot install BIOS bootloader on UEFI installation when install_hybrid_grub is 'False'!"))
-            return
+        grubinstall_command = [
+            libcalamares.job.configuration["grubInstall"],
+            "--target=" + efi_target,
+            "--efi-directory=" + efi_directory,
+            "--bootloader-id=" + efi_bootloader_id,
+            "--force"]
 
-        if libcalamares.globalstorage.value("bootLoader") is None:
+        if is_zfs:
+            # Needs environment to be set for GRUB, so go via the shell
+            check_target_env_call(["sh", "-c", "ZPOOL_VDEV_NAME_PATH=1 " + " ".join(grubinstall_command)])
+        else:
+            check_target_env_call(grubinstall_command)
+
+    else:
+        if libcalamares.globalstorage.value("bootLoader") is None and install_hybrid_grub:
             efi_install_path = libcalamares.globalstorage.value("efiSystemPartition")
             if efi_install_path is None or efi_install_path == "":
                 efi_install_path = "/boot/efi"
@@ -683,9 +686,32 @@ def run_grub_install(fw_type, partitions, efi_directory, install_hybrid_grub):
                                    boot_loader_install_path])
 
 
+def show_broken_uefi_warning():
+    warning = (
+        "Due to a probably bug of your UEFI, system bootloader has been installed "
+        "but not listed in the entry list. You need to select drive on which you "
+        "install the system by yourself for default boot later."
+    )
+    title = "Bootloader installation partially broken"
+
+    subprocess.call([
+        "yad",
+        "--width",
+        "600",
+        "--title",
+        title,
+        "--image",
+        "dialog-warning",
+        "--text",
+        warning
+    ])
+
 def get_partition_drive(partition):
     devname = partition.replace("/dev/", "")
 
+    if not os.path.exists(f"/sys/class/block/{devname}/partition"):
+        return partition, 0
+
     with open(f"/sys/class/block/{devname}/partition", "r") as f:
         partition_number = f.readline().strip()
 
@@ -725,9 +751,14 @@ def add_additional_entries_limine(efi_directory, installation_root_path, fw_type
             elif fw_type != "efi" and len(partition) == 1:
                 (devname,) = partition
                 drive, partition_number = get_partition_drive(devname)
+                drive = drive.replace("/dev/", "")
+                with open(f"/sys/block/{drive}/diskseq") as f:
+                    diskseq = f.readline().strip()
+
                 config_file.write(f"//{pretty_name}\n")
                 config_file.write(f"\tprotocol: bios_chainload\n")
-                config_file.write(f"\tdrive: {drive}\n")
+                if diskseq:
+                    config_file.write(f"\tdrive: {diskseq}\n")
                 config_file.write(f"\tpartition: {partition_number}\n")
 
 def update_limine_config(efi_directory, installation_root_path, fw_type):
@@ -773,11 +804,14 @@ def update_limine_config(efi_directory, installation_root_path, fw_type):
         except KeyError:
             libcalamares.utils.warning('limineSplashLogo not set. Skipping wallpaper in limine.conf')
 
+        machine_id = get_machine_id(installation_root_path)
+        config_file.write(f"comment: machine-id={machine_id}\n")
         config_file.write(f"/+CachyOS\n")
 
     with open(limine_entry_config_path, 'w') as limine_entry_config:
         limine_entry_config.write(f'ESP_PATH="{efi_directory}"\n')
         limine_entry_config.write(f'KERNEL_CMDLINE[default]+="{kernel_params}"\n')
+        limine_entry_config.write(f'BOOT_ORDER="*, *lts, *fallback, Snapshots"\n')
 
 def install_limine(efi_directory, fw_type):
     """
@@ -798,46 +832,25 @@ def install_limine(efi_directory, fw_type):
     if not os.path.isdir(install_efi_directory):
         os.makedirs(install_efi_directory)
 
+    update_limine_config(efi_directory, installation_root_path, fw_type)
+    add_additional_entries_limine(efi_directory, installation_root_path, fw_type)
+
     if fw_type == "efi":
         libcalamares.utils.debug("Bootloader: limine (efi)")
 
-        # VFAT is weird, see issue CAL-385
-        install_efi_directory_firmware = (vfat_correct_case(
-            install_efi_directory,
-            "EFI"))
-        if not os.path.exists(install_efi_directory_firmware):
-            os.makedirs(install_efi_directory_firmware)
+        exit_code = target_env_call(["limine-install"])
 
-        # there might be several values for the boot directory
-        # most usual they are boot, Boot, BOOT
-        install_efi_boot_directory = (vfat_correct_case(
-            install_efi_directory_firmware,
-            "boot"))
-        if not os.path.exists(install_efi_boot_directory):
-            os.makedirs(install_efi_boot_directory)
+        # Retry using fallback option if initial attempt failed
+        if exit_code != 0:
+            libcalamares.utils.warning(_("Failed to install limine bootloader, trying fallback option"))
 
-        efi_file_source = installation_root_path + "/usr/share/limine/BOOTX64.EFI"
-        efi_file_destination = os.path.join(install_efi_boot_directory, "BOOTX64.EFI")
-        shutil.copy2(efi_file_source, efi_file_destination)
+            check_target_env_call(["limine-install", "--skip-uefi", "--fallback"])
+            show_broken_uefi_warning()
 
-        devname = next(
-            partition['device'] for partition in partitions
-            if partition["mountPoint"] == efi_directory
-        )
-        drive, partition_number = get_partition_drive(devname)
-
-        # Add limine boot entry via efibootmgr
-        subprocess.call([
-            libcalamares.job.configuration["efiBootMgr"],
-            "-c",
-            "-w",
-            "-L", efi_label(efi_directory),
-            "-d", drive,
-            "-p", partition_number,
-            "-l", "\\EFI\\BOOT\\BOOTX64.EFI"
-        ])
-
-        efi_boot_next()
+            config_path = os.path.join(installation_root_path + "/etc/default", "limine")
+            with open(config_path, 'a') as config:
+                config.write("SKIP_UEFI=yes\n")
+                config.write("ENABLE_LIMINE_FALLBACK=yes\n")
     else:
         libcalamares.utils.debug("Bootloader: limine (bios)")
 
@@ -851,11 +864,8 @@ def install_limine(efi_directory, fw_type):
 
         bios_sys_source = installation_root_path + "/usr/share/limine/limine-bios.sys"
         shutil.copy2(bios_sys_source, install_efi_directory)
-        check_target_env_call(["limine", "bios-install", boot_loader["installPath"]])
-        update_limine_config(efi_directory, installation_root_path, fw_type)
-
-    update_limine_config(efi_directory, installation_root_path, fw_type)
-    add_additional_entries_limine(efi_directory, installation_root_path, fw_type)
+        (drive, _) = get_partition_drive(boot_loader["installPath"])
+        check_target_env_call(["limine", "bios-install", drive])
 
 
 def install_grub(efi_directory, fw_type, install_hybrid_grub):
